#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <yubihsm.h>

#define DEFAULT_CONNECTOR_URL "http://127.0.0.1:12345"

const char *key_label = "label";
const uint8_t password[] = "password";
const uint8_t data[] = "sudo make me a sandwich";

int main(void) {
  yh_connector *connector = NULL;
  uint8_t context[YH_CONTEXT_LEN];
  yh_session *session = NULL;
  yh_rc yrc = YHR_GENERIC_ERROR;

  uint16_t keyset = 1;

  yrc = yh_init();
  assert(yrc == YHR_SUCCESS);

  yrc = yh_init_connector(DEFAULT_CONNECTOR_URL, &connector);
  assert(yrc == YHR_SUCCESS);

  yrc = yh_connect_best(&connector, 1, NULL);
  assert(yrc == YHR_SUCCESS);

  yrc = yh_create_session_derived(connector, keyset, password, sizeof(password),
                                  false, context, sizeof(context), &session);
  assert(yrc == YHR_SUCCESS);

  yrc = yh_authenticate_session(session, context, sizeof(context));
  assert(yrc == YHR_SUCCESS);

  uint8_t session_id;
  yrc = yh_get_session_id(session, &session_id);
  assert(yrc == YHR_SUCCESS);

  printf("Successfully established session %02d using keyset %04x\n",
         session_id, keyset);

  yh_capabilities capabilities = {{0}};
  yrc = yh_capabilities_to_num("hmac_data:hmac_verify", &capabilities);
  assert(yrc == YHR_SUCCESS);

  uint16_t domain_five = 0;
  yrc = yh_parse_domains("5", &domain_five);
  assert(yrc == YHR_SUCCESS);

  uint16_t key_id = 0; // ID 0 lets the device generate an ID
  yrc = yh_util_generate_key_hmac(session, &key_id, key_label, domain_five,
                                  &capabilities, YH_ALGO_HMAC_SHA256);
  assert(yrc == YHR_SUCCESS);

  printf("Generated HMAC-SHA256 key with ID %04x\n", key_id);

  uint8_t hmac_data[64];
  size_t hmac_data_len = sizeof(hmac_data);
  yrc = yh_util_hmac(session, key_id, data, sizeof(data) - 1, hmac_data,
                     &hmac_data_len);
  assert(yrc == YHR_SUCCESS);

  printf("HMAC of data (%zu bytes) is:", hmac_data_len);
  for (uint16_t i = 0; i < hmac_data_len; i++) {
    printf(" %02x", hmac_data[i]);
  }
  printf("\n");

  bool verified;
  yrc = yh_util_hmac_verify(session, key_id, hmac_data, hmac_data_len, data,
                            sizeof(data) - 1, &verified);
  assert(yrc == YHR_SUCCESS);

  if (verified == true) {
    printf("Successfully verified HMAC\n");
  } else {
    printf("Unable to verify HMAC\n");
  }

  hmac_data[0] += 1;
  yrc = yh_util_hmac_verify(session, key_id, hmac_data, hmac_data_len, data,
                            sizeof(data) - 1, &verified);
  assert(yrc == YHR_SUCCESS);

  if (verified == true) {
    printf("Successfully verified HMAC\n");
  } else {
    printf("Unable to verify HMAC\n");
  }

  yrc = yh_util_close_session(session);
  assert(yrc == YHR_SUCCESS);

  yrc = yh_destroy_session(&session);
  assert(yrc == YHR_SUCCESS);

  yh_disconnect(connector);
  assert(yrc == YHR_SUCCESS);

  yrc = yh_exit();
  assert(yrc == YHR_SUCCESS);

  return 0;
}
